// Compile with the following for the case where KDL is entered:
//    g++-x86 -DENTER_KDL -lmidi2 midi_kdl_repro.cpp -o midi_kdl_repro
//
// Compile it without -DENTER_KDL for the case where it works, i.e. the workaround is in place.
// See `void sendMessage()' for more info.

#include <iostream>
#include <chrono>
#include <thread>

#include <MidiRoster.h>
#include <MidiConsumer.h>
#include <MidiProducer.h>


BMidiConsumer* gMidiOutConsumer;
BMidiLocalProducer* gMidiOutLocProd;

void printUsage(char* programName)
{
	bool OK = true;
	int32 id = 0;

	std::cout << "Usage: " << programName << " ID" << std::endl
	          << std::endl
	          << "Where ID is one of the following device IDs:" << std::endl;

	while (OK) {
		BMidiConsumer* c = BMidiRoster::NextConsumer(&id);
		if (c != NULL) {
			std::cout << "\t" << id << "\t" << c->Name() << std::endl;
			c->Release();
		} else {
			OK = false;
		}
	}

	std::cout << std::endl << "If no devices are shown, ensure that at least one MIDI out device"
	          << " is connected." << std::endl;
}

void sendMessage(uint32 msg)
{
	char buf[3];
	buf[0] = msg & 0xFF;
	buf[1] = (msg >> 8) & 0xFF;
	buf[2] = (msg >> 16) & 0xFF;

	std::cout << "sendMessage: " << std::hex << ((int)buf[0] & 0xFF) << " "
	                             << std::hex << (int)buf[1] << " "
	                             << std::hex << (int)buf[2] << std::endl;

#ifdef ENTER_KDL
	gMidiOutLocProd->SprayData((void*)&buf, 3, true);
		// length is always 3 bytes, which KDLs on 2 byte messages such as ProgramChange (0xC0)
		// or ChannelPressure (0xD0)
#else
	gMidiOutLocProd->SprayData((void*)&buf, (buf[0] & 0xFF) == 0xC0 ? 2 : 3, true);
		// length will be 2 bytes for ProgramChange (0xC0) messages, 3 otherwise -- therefore this
		// would still enter KDL on ChannelPressure (0xD0) messages which are also 2 bytes long.
#endif
}

void runTests()
{
	std::cout << "Sending a NoteOn message (3 bytes)" << std::endl;
	sendMessage(0x007f4090);
	std::cout << "waiting 1 second until sending the next message..." << std::endl;
	std::this_thread::sleep_for(std::chrono::seconds(1));

	std::cout << "Sending a NoteOff message (3 bytes)" << std::endl;
	sendMessage(0x00004080);
	std::cout << "waiting 2 seconds until sending the next message..." << std::endl;
	std::this_thread::sleep_for(std::chrono::seconds(2));

	std::cout << "Sending a ProgramChange message (2 bytes";
#ifdef ENTER_KDL
	std::cout << ", but 3 bytes sent -- enjoy KDL ;-)";
#endif
	std::cout << ")" << std::endl;
	sendMessage(0x000001C0);

#ifdef ENTER_KDL
	std::cout << "waiting 30 more seconds..." << std::endl;
	std::this_thread::sleep_for(std::chrono::seconds(30));
#endif
}

int main(int argc, char** argv)
{
	int32 id = 0;
	gMidiOutConsumer = NULL;
	gMidiOutLocProd = NULL;

	if (argc != 2) {
		printUsage(argv[0]);
		return 1;
	}

	try {
		id = std::stoi(argv[1]);
	} catch (std::exception const &e) {
		std::cerr << "error: could not parse the provided ID: " << argv[1]
		          << ", ensure you're passing a number" << std::endl;
		printUsage(argv[0]);
		return 2;
	}

	std::cout << "attempting to connect to device ID " << id << std::endl;

	gMidiOutConsumer = BMidiRoster::FindConsumer(id);
	if (gMidiOutConsumer == NULL) {
		std::cerr << "error: could not find consumer for the provided ID: " << id << std::endl;
		printUsage(argv[0]);
		return 3;
	}
	gMidiOutLocProd = new BMidiLocalProducer("midi_kdl_repro");
	if (!gMidiOutLocProd->IsValid()) {
		gMidiOutLocProd->Release();
		std::cerr << "error: producer is not valid" << std::endl;
		printUsage(argv[0]);
		return 4;
	}
	gMidiOutLocProd->Register();
	if (gMidiOutLocProd->Connect(gMidiOutConsumer) != B_OK) {
		std::cerr << "error: could not connect our local producer with the consumer" << std::endl;
		printUsage(argv[0]);
		return 5;
	}
	std::cout << "connected to " << id << std::endl;

	runTests();

	std::cout << "cleanup ..." << std::endl;
	gMidiOutLocProd->Disconnect(gMidiOutConsumer);
	gMidiOutConsumer->Release();
	gMidiOutLocProd->Unregister();
	gMidiOutLocProd->Release();

	return 0;
}
